﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using NAudio.Wave;
using NAudio.Wave.SampleProviders;
using System.Diagnostics;

namespace ERR_AudioPlayer
{
    /// <summary>
    /// Handles Audio Playback
    /// </summary>
    public class AudioPlayer : IAudioPlayer
    {
        private IWavePlayer waveOut = null; //Playback state reporting + Play and Stop commands
        private WaveStream fileWaveStream = null; //Stream seeking and position/duration reporting
        private SampleChannel waveChannel = null; //Playback audio volume adjustments

        private Guid lastDeviceGuid; //Last used device GUID will be stored here. Used for GUI frontend only

        private static AudioLevel audioLevel = new AudioLevel(); //Used for audio level reporting

        /// <summary>
        /// Stores current audio level
        /// </summary>
        public AudioLevel level
        {
            get
            {
                if (waveOut != null)
                {
                    try
                    {
                        return audioLevel;
                    }
                    catch (Exception ex)
                    {
                        //Redirect to your preferred logger
                        //log.Write("ERROR GetAudioLevel: " + ex.ToString());
                    }
                }

                return new AudioLevel { left = 0, right = 0 }; //Fake audio level for GUI
            }
        }
        

        public AudioPlayer(string fullFilePath, TimeSpan tcIn, Guid device, int latency, bool play)
        {
            initDevice(fullFilePath, tcIn, device, latency);
            if (play) this.Play();
        }

        
        /// <summary>
        /// DirectSound audio devices
        /// </summary>
        /// <returns>All usable DirectSound audio devices</returns>
        public static List<SoundDevice> getDevices()
        {
            List<SoundDevice> devices = new List<SoundDevice>();

            var dxAudioDevices = DirectSoundOut.Devices.ToList();
            foreach (var dxDev in dxAudioDevices)
            {
                devices.Add(new SoundDevice() { guid = dxDev.Guid, description = dxDev.Description });
            }

            return devices;
        }


        /// <summary>
        /// Makes a time jump in the playback
        /// </summary>
        /// <param name="tcIn">Timespean from the begginning of the file</param>
        /// <returns>True: Success, False: Failed</returns>
        public bool SetPosition(TimeSpan tcIn)
        {
            if (fileWaveStream != null && tcIn < fileWaveStream.TotalTime.Duration())
            {
                fileWaveStream.CurrentTime = tcIn;
                return true;
            }
            
            return false;
        }


        /// <summary>
        /// Playback timespan from the beginning of the file to the current position being played
        /// </summary>
        public TimeSpan CurrentPosition
        {
            get
            {
                if (fileWaveStream != null) return fileWaveStream.CurrentTime;

                return new TimeSpan();
            }
        }

        /// <summary>
        /// Total duration of the audio stream
        /// </summary>
        public TimeSpan Duration
        {
            get
            {
                if (fileWaveStream != null) return fileWaveStream.TotalTime;

                return new TimeSpan();
            }
        }

        /// <summary>
        /// Returns true if file is playing
        /// </summary>
        public bool IsPlaying
        {
            get
            {
                if (waveOut == null) return false;
                else if (waveOut.PlaybackState == PlaybackState.Playing)
                {
                    return true;
                }
                
                return false;
            }
        }

        /// <summary>
        /// Current volume setting of the player
        /// </summary>
        public float Volume
        {
            get
            {
                if (waveChannel != null) return waveChannel.Volume;
                else return -1;
            }
            set
            {
                if (waveChannel != null)
                {
                    if(value >=0 && value <= 1.5F) waveChannel.Volume = value; //We allow only 50% boost in volume.
                    else
                    {
                        //Redirect to your preferred logger
                        //log.Write("WARNING: Invalid volume level(" + value.ToString() + ") received. Level ignored");
                    }
                }
            }
        }

        /// <summary>
        /// Call only if you have initialized the class with play = false
        /// </summary>
        /// <returns></returns>
        public bool Play()
        {
            if (waveOut != null)
            {
                waveOut.Play();
                return true;
            }
            else return false;
        }
        
        /// <summary>
        /// Stop playback
        /// </summary>
        /// <returns>Always True</returns>
        public bool Stop()
        {
            if (waveOut != null && waveOut.PlaybackState != PlaybackState.Stopped)
            {
                waveOut.Stop();
            }

            return true;
        }

        /// <param name="hours"></param>
        /// <param name="minutes"></param>
        /// <param name="seconds"></param>
        /// <param name="milliseconds"></param>
        /// <returns>True: Operation Successful, False: Unable to seek</returns>
        public bool SeekForward(int hours, int minutes, int seconds, int milliseconds)
        {
            if (fileWaveStream != null)
            {
                var result = fileWaveStream.CurrentTime.Add(new TimeSpan(0, hours, minutes, seconds, milliseconds));

                if (result < fileWaveStream.TotalTime)
                {
                    fileWaveStream.CurrentTime = result;
                    return true;
                }
            }

            return false;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="hours"></param>
        /// <param name="minutes"></param>
        /// <param name="seconds"></param>
        /// <param name="milliseconds"></param>
        /// <returns>True: Operation Successful, False: Unable to seek</returns>
        public bool SeekBackward(int hours, int minutes, int seconds, int milliseconds)
        {
            if (fileWaveStream != null)
            {
                var result = fileWaveStream.CurrentTime.Subtract(new TimeSpan(0, hours, minutes, seconds, milliseconds));

                if (result.Ticks > 0)
                {
                    fileWaveStream.CurrentTime = result;
                    return true;
                }
            }

            return false;
        }

        /// <returns>Last used device GUID or null when no device has been used yet</returns>
        public Guid? GetDevice()
        {
              return lastDeviceGuid;
        }

        /// <summary>
        /// We open the actual file here. If needed WaveFormat Encoder is added to the audio pipeline.
        /// </summary>
        /// <param name="fileFullPath"></param>
        /// <param name="tcIn"></param>
        /// <returns>Operational WaveStream with the needed converters attached</returns>
        private WaveStream GetWaveStream(string fileFullPath, TimeSpan tcIn)
        {
            WaveStream readerStream = new WaveFileReader(fileFullPath);

            if (readerStream.WaveFormat.Encoding != WaveFormatEncoding.Pcm && readerStream.WaveFormat.Encoding != WaveFormatEncoding.IeeeFloat)
            {
                //Redirect to your preferred logger
                //log.Write("WARNING: Wave format conversion in use: Encoding: " + readerStream.WaveFormat.Encoding.ToString() + " File path: " + fileFullPath);

                readerStream = WaveFormatConversionStream.CreatePcmStream(readerStream);
                readerStream = new BlockAlignReductionStream(readerStream);
            }

             //Set audio inpoint
            if(tcIn < readerStream.CurrentTime.Duration()) readerStream.CurrentTime = tcIn;
            return readerStream;
        }

        /// <summary>
        /// Add audio sample metering to the pipeline
        /// </summary>
        private MeteringSampleProvider GetMeteringSampleProvider(ISampleProvider source)
        {
            var postVolumeMeter = new MeteringSampleProvider(source);
            postVolumeMeter.StreamVolume += postVolumeMeter_StreamVolume;

            return postVolumeMeter;
        }

        /// <summary>
        /// Configures all the needed streams. Selects the device and seeks to the desired position in the file.
        /// </summary>
        /// <param name="fileFullPath">Full directory path to the file + filename with extension</param>
        /// <param name="tcIn">Starting position of the playback. If tcIn > duration tcIn = "00:00:00:000"</param>
        /// <param name="device">Audio device GUID to use for the playback</param>
        /// <param name="latency">50 for most systems. Increase the value when playback is unstable or crashes the soundcard drivers</param>
        private bool initDevice(string fileFullPath, TimeSpan tcIn, Guid device, int latency)
        {
            if (waveOut != null || fileWaveStream != null)
            {
                Stop();
                waveOut = null;
                fileWaveStream = null;
            }

            //Initialize playout device
            try
            {
                waveOut = new DirectSoundOut(device, latency);
                waveOut.PlaybackStopped += waveOut_PlaybackStopped;

                lastDeviceGuid = device; //Used for GUI to show the "current device" selection
            }
            catch (Exception driverCreateException)
            {
                //Redirect to your preferred logger
                //log.Write("ERROR: Unable to initialize DirectSoundOut device: " + device + String.Format("{0}", driverCreateException.Message));
                return false;
            }

            //Initialize streams
            try
            {
                fileWaveStream = GetWaveStream(fileFullPath, tcIn);
                waveChannel = new SampleChannel(fileWaveStream);

                initWaveOut(GetMeteringSampleProvider(waveChannel));
            }
            catch (Exception ex)
            {
                //Redirect to your preferred logger
                //log.Write("ERROR: Unable to create streams" + ex.ToString());
                return false;
            }

            return true;
        }

        private bool initWaveOut(ISampleProvider sampleProvider)
        {
            //Initialize Wave Output
            try
            {
                //Redirect to your preferred logger
                //log.Write("RATE: " + sampleProvider.WaveFormat.SampleRate.ToString());
                waveOut.Init(new SampleToWaveProvider(sampleProvider));
                
                return true;
            }
            catch (Exception ex)
            {
                //Redirect to your preferred logger
                //log.Write("ERROR: Unable to initialize Wave output!" + ex.ToString());
                return false;
            }
        }
       
        void waveOut_PlaybackStopped(object sender, StoppedEventArgs e)
        {
            if (waveOut != null && waveOut.PlaybackState != PlaybackState.Playing)
            {
                //Redirect to your preferred logger
                //log.Write("Releasing resources");
                waveOut.Dispose();
                waveOut = null;
                if (fileWaveStream != null) fileWaveStream.Close();
                level.left = 0;
                level.right = 0;
            }
        }

        private void postVolumeMeter_StreamVolume(object sender, StreamVolumeEventArgs e)
        {
            if (e.MaxSampleValues.Length == 2)
            {
                audioLevel.left = e.MaxSampleValues[0];
                audioLevel.right = e.MaxSampleValues[1];
            }
           
            audioLevel.left = e.MaxSampleValues[0];
            audioLevel.right = audioLevel.left; 
        }
    }
}
